-- Insertion sort example for the quickCheck plugin.
-- Author: Ulf Norell
-- Created: 2004-12-09

-- # import QuickCheck

-- Insertion sort ---------------------------------------------------------

Leq : Set -> Set
Leq A = A -> A -> Bool

insert (A : Set) : Leq A -> A -> List A -> List A
insert _ x [] = [x]
insert (<=) x (y :: ys) =
    if (x <= y)
       (x :: y :: ys)
       (y :: insert (<=) x ys)

isort (A : Set) : Leq A -> List A -> List A
isort (<=) []	     = []
isort (<=) (x :: xs) = insert (<=) x (isort (<=) xs)

-- Testing ----------------------------------------------------------------

infix 12 <=
(<=) : Leq Nat
0     <= _     = true
suc n <= 0     = false
suc n <= suc m = n <= m

-- Helpers
ordered : List Nat -> Bool
ordered []	   = true
ordered [x]	   = true
ordered (x :: y :: xs) = x <= y && ordered (y :: xs)

-- Generators

-- Ordered lists
orderedGen : Gen (List Nat)
orderedGen = liftM monadGen (scanl1 (+)) (arbList arbNat)

-- Properties
PropInsertBad : Nat -> List Nat -> Prop
PropInsertBad x xs =
    Holds (ordered xs) ==> 
      collect (length xs)
	(Holds (ordered (insert (<=) x xs)))

test_insBad : (x : Nat) -> (xs : List Nat) -> Tested (PropInsertBad x xs)
test_insBad = qc-plugin(arbNat, arbList arbNat)

-- We don't have to use forAll here, but since we are relying on the
-- fact that the list is ordered it's good to specify the generator
-- in the property and not leave it up to the tester.
PropInsertGood : Nat -> Prop
PropInsertGood x =
    forAll orderedGen $ \xs ->
    collect (length xs)
    $ Holds $ ordered (insert (<=) x xs)

test_insGood : (x : Nat) -> Tested (PropInsertGood x)
test_insGood = qc-plugin(arbNat)

PropInsertLength : Nat -> List Nat -> Prop
PropInsertLength x xs =
    suc (length xs) == length (insert (<=) x xs)

test_insLength : (x : Nat) -> (xs : List Nat) -> Tested (PropInsertLength x xs)
test_insLength = qc-plugin(arbNat, arbList arbNat)

