-- #import Framework

-- The extensional axiom of choice
ExtAC (A,B : Set) : (R : A -> B -> Set) -> 
          (eqA : Rel A) -> (eqB : Rel B) -> 
          EquivRel eqA -> EquivRel eqB -> 
          ((x,y : A) -> x `eqA` y -> (z : B) -> R x z -> R y z) ->
          ((x,y : B) -> x `eqB` y -> (z : A) -> R z x -> R z y) -> 
          ((x : A) -> Exists (\y -> R x y)) ->
          Exists (\f -> 
                  ((x : A) -> R x (f x)) /\ 
                  ((x,y : A) -> x `eqA` y -> (f x) `eqB` (f y)))

-- a logical tautology we need
taut1 (A,B,C,D : Set) : A \/ B -> (A -> C) -> (B -> D) -> C \/ D
taut1 = \h h1 h2 -> orE h 
        (\a -> orI1 (h1 a ))
        (\b -> orI2 (h2 b ))

-- and an other one, constraposition
contrap (A,B : Set) : (A -> B) -> Not B -> Not A
contrap h1 h2 = \a -> h2 (h1 a)

-- an equivalence relation on Bool
eqb : Rel Bool            
eqb false false = True
eqb false true = False
eqb true false = False
eqb true true = True

R_eqb : Reflexive eqb
R_eqb true = tt
R_eqb false = tt

S_eqb : Symmetrical eqb
S_eqb false false h = tt
S_eqb false true h = h
S_eqb true false h = h
S_eqb true true h = tt

T_eqb : Transitive eqb
T_eqb false false false h1 h2 = tt
T_eqb false false true h1 h2 = h2
T_eqb false true false h1 h2 = tt
T_eqb false true true h1 h2 = h1
T_eqb true false false h1 h2 = h1
T_eqb true false true h1 h2 = tt
T_eqb true true false h1 h2 = h2
T_eqb true true true h1 h2 = tt

Equiv_eqb : EquivRel eqb
Equiv_eqb = (R_eqb, S_eqb, T_eqb)

-- eqb is decidable
Decid_eqb : (x,y : Bool) -> eqb x y \/ Not (eqb x y)
Decid_eqb false false = inl tt
Decid_eqb false true = inr (\ff -> ff)
Decid_eqb true false = inr (\ff -> ff)
Decid_eqb true true = inl tt

-- the relation R
R' : (P : Set) -> Rel Bool
R' P a b = eqb a b \/ P

R_R' : (P : Set) -> Reflexive (R' P)
R_R' P x = inl (R_eqb x)

S_R' : (P : Set) -> Symmetrical (R' P)
S_R' P x1 x2 h = orE h
                 (\heqb -> inl (S_eqb x1 x2 heqb))
                 (\hP -> inr hP)

T_R' : (P : Set) -> Transitive (R' P)
T_R' P x1 x2 x3 h1 h2 =
    orE h1
    (\heqb12 -> orE h2 
                (\heqb23 -> inl (T_eqb x1 x2 x3 heqb12 heqb23))
                (\hp -> inr hp))
    (\hp     -> inr hp)

EquivR' : (P : Set) -> EquivRel (R' P)
EquivR' P = (R_R' P, S_R' P, T_R' P)

LEM : (P : Set) -> P \/ Not P
LEM P = let R = R' P
            EquivR = EquivR' P
            trans = EquivR.2.2
            symm = EquivR.2.1
            refl = EquivR.1

            -- R is left - extensional
            LExtR : (x,y : Bool) -> R x y -> (z : Bool) -> R x z -> R y z
            LExtR x y h1 z h2 = trans y x z (symm x y h1) h2

            -- R is right - extensional
            RExtR : (x,y : Bool) -> eqb x y -> (z : Bool) -> R z x -> R z y
            RExtR x y h1 z h2 = trans z x y h2 (orI1 h1)

            -- use ExtAC to get the choice function
            choicef : Exists
                        (\f -> 
                             (((x : Bool) -> R x (f x)) /\ 
                              ((x,y : Bool) -> x `R` y -> (f x) `eqb` (f y))))
            choicef = let lemma : (x : Bool) -> Exists (\y -> R x y)
                          lemma = \x -> (x, (refl x ))
                      in ExtAC R R eqb EquivR Equiv_eqb LExtR RExtR lemma

            -- extract the function only
            cf : Bool -> Bool
            cf = choicef.1

            -- decidability, direction 1
            decid1 : eqb (cf false) (cf true) -> P
            decid1 =
                let
                 lemma7 : R false true -> P
                 lemma7 h = orE h 
                                (\ff -> whenFalse P ff )
                                (\p -> p )
                 lemma5 : (x,y : Bool) -> eqb (cf x) (cf y) -> R x y
                 lemma5 x y h = 
                     let
                      step1 : R x (cf x)
                      step1 = choicef.2.1 x
                      step2 : R x (cf y)
                      step2 = RExtR (cf x )
                                    (cf y )
                                    h
                                    x 
                                    (choicef.2.1 x)
                      step3 : R y (cf y)
                      step3 = choicef.2.1 y
                      step4 : R (cf y) y
                      step4 = symm y (cf y) step3 
                     in trans x (cf y) y step2 step4
                in \h -> lemma7 (lemma5 false true h) 

            -- decidability, direction 2
            decid2 : P -> eqb (cf false) (cf true)
            decid2 =
                let lemma8 : P -> R false true
                    lemma8 = orI2
                    lemma6 : (x,y : Bool) -> R x y -> eqb (cf x) (cf y)
                    lemma6 = choicef.2.2
                in \h -> lemma6 false true (lemma8 h)

        in taut1 (Decid_eqb (cf false) (cf true)) decid1 (contrap decid2)
