{-# OPTIONS -fallow-overlapping-instances -fglasgow-exts #-}
{-| Abstracting a value @w@ from another value @v@ is the process of
    finding a function @f@ such that @f w = v@.
-}

module TypeChecker.Abstract where

import Control.Monad.State

import Syntax.Internal
import Syntax.Values

import TypeChecker.Monad
import TypeChecker.UnfoldConstants
import TypeChecker.Universes
import TypeChecker.Values
import TypeChecker.Abstractions

import Utils.Monads

-- * Abstraction

class Abstract v where
    abstr :: Value -> v -> [v]

-- | @abstract w v@, abstract w from v the result is a list of values with one extra
--   free variable (the hole).
abstract :: Value -> Value -> TC [Value]
abstract w v =
    do	w' <- unfoldAllConstants w
	v' <- unfoldAllConstants v
	return $ abstr w' v'

instance Abstract Value where
    abstr w v
	| w == v	= [ vBId 0, raise v ]
	| otherwise	=
	    case vView v of
		VId _	     -> return $ raise v
		VBId _	     -> return $ raise v
		VDef _ _     -> return $ raise v
		VCon _ _ _ _ -> return $ raise v
		VTop	     -> return $ raise v
		VMeta _	     -> return $ raise v
		VSort _	     -> return $ raise v
		VApp v v'    -> abstr2 vApp w v v'
		VPair v v'   -> abstr2 vPair w v v'
		VPi v f	     -> abstr2 vPi w v f
		VLam f	     -> vLam <$> abstr w f
		VProj v p    -> abstr2 vProj w v p
		VSigma v f   -> abstr2 vSigma w v f
		VStruct xs   -> vStruct <$> abstr w xs

abstr2 f w x y = liftM2 f (abstr w x) (abstr w y)

instance (Abstract a, Abstract b) => Abstract (a,b) where
    abstr w (x,y) = abstr2 (,) w x y

instance Abstract a => Abstract [a] where
    abstr w xs = mapM (abstr w) xs

instance Abstract String where
    abstr w s = [s]

instance Abstract Int where
    abstr w n = [n]

instance Abstract ProjectionValue where
    abstr w p = [p]

instance Abstract Type where
    abstr w t = typeMapM (abstr w) t

instance (Instantiate v, Raiseable v, Abstract v) => Abstract (Abs v) where
    abstr w f = absMapM (map swap . abstr (raise w)) f
	where
	    swap v = instantiateUnder 2 (raiseBy 2 v) [vBId 1, vBId 0]

